#! /usr/bin/perl -w
use strict;
use utf8;

#############################################
#
#	file:	AddYomiToIdx.pl
#
#	project:	JTeXEdit
#
#	Copyright (c) 1999-2006 Norihito KOISO
#	All Rights Reserved
#
#############################################
#
#	使用法：
#		LaTeX が生成した "***.idx" を AddYomiToIdx.pl に落とせば
#		漢字に読みを付けた "***.idxy" を生成する．
#
#		辞書ファイル "***.idxd" を生成するので読みを修正できる．
#
#############################################
#
#	履歴
#	2011-04-03: use Foundation
#	2008-06-05: perl UTF-8 環境 に対応
#	2006-04-29: JLAA ではなく JTeXEdit で制御
#	2003-10-11: AppleScript の不具合を修正
#	2003-06-15: shell perl (+JLAA) 専用
#	2002-07-15: unix 改行形式に対応
#	2002-01-26:	初版
#
#############################################

#use encoding 'utf8'; 
use open IO => ':encoding(shift_jis)';
use Foundation;
# man PerlObjCBridge

sub SetFileTypes($) {
	my ($path) = @_;
	my $sign = NSNumber->numberWithInt_(unpack("N", 'jtxE'));
	my $type = NSNumber->numberWithInt_(unpack("N", 'TEXT'));
	my $dict = NSMutableDictionary->dictionary;
	$dict->setObject_forKey_($sign, "NSFileHFSCreatorCode");
	$dict->setObject_forKey_($type, "NSFileHFSTypeCode");
	NSFileManager->defaultManager
		->changeFileAttributes_atPath_($dict, $path);
}

my %Dict = ();
my $DictIsDirty = '';
sub ReadDict() {
	$DictIsDirty = '';
	%Dict = ();
	while (defined(my $line = <IN>)) {
		chomp $line;
		my @items = split(/\n|\r/, $line);
		for my $item (@items) {
			next if ($item =~ /^%/ or $item eq '');
			my @content = split(/\t/, $item);
			if (@content == 2) {
				$Dict{$content[1]} = $content[0];
			} else {
				warn "辞書の形式が正しくありません:\n$item\n";
			}
		}
	}
}
sub WriteDict() {
	my %buff = ();
	while (my ($text, $yomi) = each %Dict) {
		$buff{"$yomi\t$text"} = 1;
	}
	for my $item (sort keys %buff) {
		print OUT "$item\n";
	}
}

sub DoAppleScript($;$) {
	my ($text, $here) = @_;
	$here = 'HERE_DOC' unless ($here);
	return `/usr/bin/osascript -ss <<'$here'\n$text\n$here`;
}

sub Split_Kanji_Kana($) {
	my ($text) = @_;
	my @out = ();
	while (length $text) {
		$text =~ s/^([^ァ-ヶー－ぁ-ゖ]*)//;
		push(@out, $1);
		$text =~ s/^([ァ-ヶー－ぁ-ゖ]*)//;
		push(@out, $1);
	}
	return @out;
}

sub CheckDict($$);
sub CheckDict($$) {
	my ($text, $kanaLen) = @_;
	my $textLen = length($text);
	return $text if ($textLen <= $kanaLen);
	my $yomi = $Dict{$text};
	return $yomi if (defined($yomi));
	for (my $i = $textLen - 1; $i > 0; --$i) {
		my $left = substr($text, 0, $i);
		my $yomiL = $Dict{$left};
		next unless $yomiL;
		my $yomiR = CheckDict(substr($text, $i), $kanaLen);
		return "$yomiL$yomiR" if (defined($yomiR));
	}
	return undef;
}

sub CallJLAA(@) {
	my @list = @_;
	my $listtext = '{"' . join('", "', @list) . '"}';
	my $script =<<EOS;
set theList to $listtext
tell application "JTeXEdit" to LAAConvert theList convType ToHiragana
EOS
	my $result = DoAppleScript($script);
	$result =~ s/^["{\r\n]*//;
	$result =~ s/["}\r\n]*$//;
	return split(/", "/, $result);
}

my %Unknowns;
my $SavingUnknowns;

sub AskJLAA() {
	my @unknowns = sort keys %Unknowns;
	undef %Unknowns;
	$SavingUnknowns = '';
	return unless @unknowns;
	my @yomilist = CallJLAA(@unknowns);
	if (@yomilist != @unknowns) {
		for (my $i = 0; $i <= $#unknowns; ++$i) {
			print "'$unknowns[$i]' => '$yomilist[$i]'\n";
		}
		die "JLAA の結果が不完全です．\n";
	}
	for (my $i = 0; $i <= $#yomilist; ++$i) {
		my $yomi = $yomilist[$i];
		my $text = $unknowns[$i];
		if (length($yomi)) {
			print "$text > $yomi\n";
			$DictIsDirty = 1;
			$Dict{$text} = $yomi;
		} else {
			warn "'$text' が変換できませんでした．\n";
		}
	}
}

sub CheckOneYomi($$) {
	my ($kanji, $kana) = @_;
	return $kana if ($kanji eq '');
	my $text = $kanji.$kana;
	my $yomi = CheckDict($text, length($kana));
	return $yomi if (defined($yomi));
	return undef;
}
sub MakeOneYomi($$) {
	my ($kanji, $kana) = @_;
	my $yomi = CheckOneYomi($kanji, $kana);
	return $yomi if defined($yomi);
	if ($SavingUnknowns) {
		$Unknowns{$kanji.$kana} = 1;
	} else {
	}
	return $kanji.$kana;
}

#sub MakeIgnoreChars() {
#	# SJIS 8100-81FF 句読点など;
#	# SJIS 849F-8860 掛け線 ┴ ;
#	my $out = '';
#	for my $sjis (0x8100 .. 0x81ff, 0x849f .. 0x8860) {
#		my $data = pack("n", $sjis);
#		my $char = Encode::decode('shift_jis', $data);
#		$out .= $char if (ord($char) != 0xfffd);
#	}
#	return $out;
#}
sub DeleteIgnoreChars($) {
	my ($text) = @_;
	# can't use $GlobalIgnoreChars in tr.
	$text =~ tr/、。，．・：；？！゛゜´｀¨＾￣＿ヽヾゝゞ〃仝々〆〇ー―‐／＼〜‖｜…‥‘’“”（）〔〕［］｛｝〈〉《》「」『』【】＋−±×÷＝≠＜＞≦≧∞∴♂♀°′″℃￥＄¢£％＃＆＊＠§☆★○●◎◇◆□■△▲▽▼※〒→←↑↓〓∈∋⊆⊇⊂⊃∪∩∧∨¬⇒⇔∀∃∠⊥⌒∂∇≡≒≪≫√∽∝∵∫∬Å‰♯♭♪†‡¶◯─│┌┐┘└├┬┤┴┼━┃┏┓┛┗┣┳┫┻╋┠┯┨┷┿┝┰┥┸╂//d;
	return $text;
}

sub MakeYomi($) {
	my ($text) = @_;
	return '' unless (length($text));
	$text = DeleteIgnoreChars($text);
	$text =~ tr/ァ-ヶ/ぁ-ゖ/;
	return $Dict{$text} if (defined($Dict{$text}));
	my @list = Split_Kanji_Kana($text);
	my $yomi = '';
	if ($list[0] eq '' and @list > 3) {
		my $dict = CheckDict("$list[1]$list[2]$list[3]", length($list[3]));
		if ($dict) {
			$yomi = $dict;
			splice(@list, 0, 4);
		}
	}
	while (@list) {
		my ($kanji, $kana) = splice(@list, 0, 2);
		$yomi .= MakeOneYomi($kanji, $kana);
	}
	return $yomi;
}

sub AddYomiToTerms(@) {
	my @terms = @_;
	for my $term (@terms) {
		my $hadyomi = @$term > 1;
		my $str = ($hadyomi) ? shift @$term : $$term[0];
		$str =~ tr/ //s;
		($str) = $str =~ /^ ?(.*) ?$/;
		$str =~ tr/a-z/A-Z/;
		$str = DeleteIgnoreChars($str);
		unless ($hadyomi) {
			$str =~ s/([\x{100}-\x{ffff}]+)/MakeYomi($1)/eg;
			$str =~ tr/ァ-ヶ/ぁ-ゖ/;
			$str =~ tr
<ぁぃぅぇぉがぎぐげござじずぜぞだぢっづでどばぱびぴぶぷべぺぼぽゃゅょゎゐゑゔゕゖ>
<あいうえおかきくけこさしすせそたちつつてとははひひふふへへほほやゆよわいえふかけ>;
		}
		unshift(@$term, $str);
	}
	return @terms;
}

sub LineToIdx($$@) {
	my ($cmd, $page, @terms) = @_;
	$cmd = "|$cmd" if $cmd;
	print OUT "\\indexentry\{",
				join('!', map {join('@', @$_)} @terms),
				"$cmd}{$page}\n";
}

sub SplitBody($) {
	my ($body) = @_;
	my @text = split(/\|/, $body);
	return "| が複数あります:" if (@text > 2);
	$body = shift(@text);
	my $cmd = @text ? $text[0] : '';
	return "| の後に ! があります:" if ($cmd =~ tr/!//);
	my @terms = map {[split('@', $_)]} split('!', $body);
	return "! が多すぎます:" if (@terms > 3);
	for (@terms) { return "\@ が多すぎます:" if (@$_ > 2); }
	return ('', $cmd, @terms);
}

sub ParseFile() {
	while (defined(my $line = <IN>)) {
		chomp $line;
		my $err = '行に誤りがあります:';
		if ($line =~ /^\\indexentry\{(.*)\}\{(\d+)\}$/) {
			my $page = $2;
			my ($cmd, @terms);
			($err, $cmd, @terms) = SplitBody($1);
			if ($SavingUnknowns) {
				AddYomiToTerms(@terms);
				$err = '';
			} else {
				LineToIdx($cmd, $page, AddYomiToTerms(@terms)) unless $err;
			}
		}
		warn "ERROR: $err\n$line\n" if $err;
	}
}

print "\n**** AddYomiToIdx.pl ****\n";

die ".idx ファイルを指定して下さい．\n" unless @ARGV;
my @OutFile = ();
while (my $idxFile = shift) {
	my $dictFile = $idxFile.'d';
	print "辞書読込 < $dictFile\n";
	if (-e $dictFile) {
		open(IN, "<:encoding(shift_jis)", $dictFile) or die "辞書が開けません\n";
		ReadDict();
		close(IN);
	} else {
		print "辞書はありません\n";
	}
	my $outFile = $idxFile.'y';
	print "処理開始: $idxFile\n > $outFile\n";
	open(IN, "<:encoding(shift_jis)", $idxFile) or die "ファイルが開けません:\n'$idxFile'\n";
	my $temp; # is it mac EOL, or unix EOL
	read(IN, $temp, 1000) or die;
	seek(IN, 0, 0);
	$/ = "\r" if ($temp =~ /\r/); # if unix EOL
	$SavingUnknowns = 1; # make unknowns list
	ParseFile();
	seek(IN, 0, 0);
	AskJLAA();
	$SavingUnknowns = '';
	open(OUT, ">:encoding(shift_jis)", $outFile) or die "ファイルが開けません:\n'$outFile'\n";
	ParseFile();
	close(IN);
	close(OUT);
	SetFileTypes($outFile);
	$/ = "\n"; # restore original EOL
	if ($DictIsDirty) {
		print "辞書書込 > $dictFile\n";
		open(OUT, ">:encoding(shift_jis)", $dictFile) or die "辞書に書けません\n";
		WriteDict();
		close OUT;
		SetFileTypes($dictFile);
	}
	push(@OutFile, $outFile);
}
print "終了．\n";

print join(",\t", @OutFile), "\n";

exit 1;
